
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#include "enums.h"
#include "app-filter.h"
#include "malcontent.h"
#include "manager.h"
#include "session-limits.h"
#include "user.h"
#include "user-manager.h"
#include "web-filter.h"

#define C_ENUM(v) ((gint) v)
#define C_FLAGS(v) ((guint) v)

/* enumerations from "app-filter.h" */

GType
mct_app_filter_oars_value_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (MCT_APP_FILTER_OARS_VALUE_UNKNOWN), "MCT_APP_FILTER_OARS_VALUE_UNKNOWN", "unknown" },
        { C_ENUM (MCT_APP_FILTER_OARS_VALUE_NONE), "MCT_APP_FILTER_OARS_VALUE_NONE", "none" },
        { C_ENUM (MCT_APP_FILTER_OARS_VALUE_MILD), "MCT_APP_FILTER_OARS_VALUE_MILD", "mild" },
        { C_ENUM (MCT_APP_FILTER_OARS_VALUE_MODERATE), "MCT_APP_FILTER_OARS_VALUE_MODERATE", "moderate" },
        { C_ENUM (MCT_APP_FILTER_OARS_VALUE_INTENSE), "MCT_APP_FILTER_OARS_VALUE_INTENSE", "intense" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("MctAppFilterOarsValue"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* enumerations from "manager.h" */

GType
mct_manager_get_value_flags_get_type (void)
{
    static gsize gtype_id = 0;
    static const GFlagsValue values[] = {
        { C_FLAGS (MCT_MANAGER_GET_VALUE_FLAGS_NONE), "MCT_MANAGER_GET_VALUE_FLAGS_NONE", "none" },
        { C_FLAGS (MCT_MANAGER_GET_VALUE_FLAGS_INTERACTIVE), "MCT_MANAGER_GET_VALUE_FLAGS_INTERACTIVE", "interactive" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_flags_register_static (g_intern_static_string ("MctManagerGetValueFlags"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

GType
mct_manager_set_value_flags_get_type (void)
{
    static gsize gtype_id = 0;
    static const GFlagsValue values[] = {
        { C_FLAGS (MCT_MANAGER_SET_VALUE_FLAGS_NONE), "MCT_MANAGER_SET_VALUE_FLAGS_NONE", "none" },
        { C_FLAGS (MCT_MANAGER_SET_VALUE_FLAGS_INTERACTIVE), "MCT_MANAGER_SET_VALUE_FLAGS_INTERACTIVE", "interactive" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_flags_register_static (g_intern_static_string ("MctManagerSetValueFlags"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

GType
mct_manager_error_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (MCT_MANAGER_ERROR_INVALID_USER), "MCT_MANAGER_ERROR_INVALID_USER", "invalid-user" },
        { C_ENUM (MCT_MANAGER_ERROR_PERMISSION_DENIED), "MCT_MANAGER_ERROR_PERMISSION_DENIED", "permission-denied" },
        { C_ENUM (MCT_MANAGER_ERROR_INVALID_DATA), "MCT_MANAGER_ERROR_INVALID_DATA", "invalid-data" },
        { C_ENUM (MCT_MANAGER_ERROR_DISABLED), "MCT_MANAGER_ERROR_DISABLED", "disabled" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("MctManagerError"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* enumerations from "user.h" */

GType
mct_user_type_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (MCT_USER_TYPE_UNKNOWN), "MCT_USER_TYPE_UNKNOWN", "unknown" },
        { C_ENUM (MCT_USER_TYPE_SYSTEM), "MCT_USER_TYPE_SYSTEM", "system" },
        { C_ENUM (MCT_USER_TYPE_CHILD), "MCT_USER_TYPE_CHILD", "child" },
        { C_ENUM (MCT_USER_TYPE_PARENT), "MCT_USER_TYPE_PARENT", "parent" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("MctUserType"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* enumerations from "web-filter.h" */

GType
mct_web_filter_type_get_type (void)
{
    static gsize gtype_id = 0;
    static const GEnumValue values[] = {
        { C_ENUM (MCT_WEB_FILTER_TYPE_NONE), "MCT_WEB_FILTER_TYPE_NONE", "none" },
        { C_ENUM (MCT_WEB_FILTER_TYPE_BLOCKLIST), "MCT_WEB_FILTER_TYPE_BLOCKLIST", "blocklist" },
        { C_ENUM (MCT_WEB_FILTER_TYPE_ALLOWLIST), "MCT_WEB_FILTER_TYPE_ALLOWLIST", "allowlist" },
        { 0, NULL, NULL }
    };
    if (g_once_init_enter (&gtype_id)) {
        GType new_type = g_enum_register_static (g_intern_static_string ("MctWebFilterType"), values);
        g_once_init_leave (&gtype_id, new_type);
    }
    return (GType) gtype_id;
}

/* Generated data ends here */

