/*
 * This file is generated by gdbus-codegen, do not modify it.
 *
 * The license of this code is the same as for the D-Bus interface description
 * it was derived from. Note that it links to GLib, so must comply with the
 * LGPL linking clauses.
 */

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include "extension-agent-iface.h"

#include <string.h>

/* ------------------------------------------------------------------------ */
/* Definitions for org.freedesktop.MalcontentTimer1.ExtensionAgent */

const GDBusArgInfo org_freedesktop_malcontent_timer1_extension_agent_interface__request_extension_method_in_args__record_type_arg =
{
  -1,  /* ref count */
  (gchar *) "record_type",
  (gchar *) "s",
  NULL,  /* no annotations */
};

const GDBusArgInfo org_freedesktop_malcontent_timer1_extension_agent_interface__request_extension_method_in_args__identifier_arg =
{
  -1,  /* ref count */
  (gchar *) "identifier",
  (gchar *) "s",
  NULL,  /* no annotations */
};

const GDBusArgInfo org_freedesktop_malcontent_timer1_extension_agent_interface__request_extension_method_in_args__duration_secs_arg =
{
  -1,  /* ref count */
  (gchar *) "duration_secs",
  (gchar *) "t",
  NULL,  /* no annotations */
};

const GDBusArgInfo org_freedesktop_malcontent_timer1_extension_agent_interface__request_extension_method_in_args__subject_arg =
{
  -1,  /* ref count */
  (gchar *) "subject",
  (gchar *) "(sa{sv})",
  NULL,  /* no annotations */
};

const GDBusArgInfo org_freedesktop_malcontent_timer1_extension_agent_interface__request_extension_method_in_args__extra_data_arg =
{
  -1,  /* ref count */
  (gchar *) "extra_data",
  (gchar *) "a{sv}",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const org_freedesktop_malcontent_timer1_extension_agent_interface__request_extension_method_in_args[] =
{
  &org_freedesktop_malcontent_timer1_extension_agent_interface__request_extension_method_in_args__record_type_arg,
  &org_freedesktop_malcontent_timer1_extension_agent_interface__request_extension_method_in_args__identifier_arg,
  &org_freedesktop_malcontent_timer1_extension_agent_interface__request_extension_method_in_args__duration_secs_arg,
  &org_freedesktop_malcontent_timer1_extension_agent_interface__request_extension_method_in_args__subject_arg,
  &org_freedesktop_malcontent_timer1_extension_agent_interface__request_extension_method_in_args__extra_data_arg,
  NULL,
};

const GDBusArgInfo org_freedesktop_malcontent_timer1_extension_agent_interface__request_extension_method_out_args__request_object_arg =
{
  -1,  /* ref count */
  (gchar *) "request_object",
  (gchar *) "o",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const org_freedesktop_malcontent_timer1_extension_agent_interface__request_extension_method_out_args[] =
{
  &org_freedesktop_malcontent_timer1_extension_agent_interface__request_extension_method_out_args__request_object_arg,
  NULL,
};

const GDBusMethodInfo org_freedesktop_malcontent_timer1_extension_agent_interface__request_extension_method =
{
  -1,  /* ref count */
  (gchar *) "RequestExtension",
  (GDBusArgInfo **) org_freedesktop_malcontent_timer1_extension_agent_interface__request_extension_method_in_args,
  (GDBusArgInfo **) org_freedesktop_malcontent_timer1_extension_agent_interface__request_extension_method_out_args,
  NULL,  /* no annotations */
};

const GDBusMethodInfo * const org_freedesktop_malcontent_timer1_extension_agent_interface_methods[] =
{
  &org_freedesktop_malcontent_timer1_extension_agent_interface__request_extension_method,
  NULL,
};

const GDBusInterfaceInfo org_freedesktop_malcontent_timer1_extension_agent_interface =
{
  -1,  /* ref count */
  (gchar *) "org.freedesktop.MalcontentTimer1.ExtensionAgent",
  (GDBusMethodInfo **) org_freedesktop_malcontent_timer1_extension_agent_interface_methods,
  NULL,  /* no signals */
NULL,  /* no properties */
  NULL,  /* no annotations */
};

