/*
 * This file is generated by gdbus-codegen, do not modify it.
 *
 * The license of this code is the same as for the D-Bus interface description
 * it was derived from. Note that it links to GLib, so must comply with the
 * LGPL linking clauses.
 */

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include "extension-agent-request-iface.h"

#include <string.h>

/* ------------------------------------------------------------------------ */
/* Definitions for org.freedesktop.MalcontentTimer1.ExtensionAgent.Request */

const GDBusMethodInfo org_freedesktop_malcontent_timer1_extension_agent_request_interface__close_method =
{
  -1,  /* ref count */
  (gchar *) "Close",
  NULL,  /* no in args */
  NULL,  /* no out args */
  NULL,  /* no annotations */
};

const GDBusMethodInfo * const org_freedesktop_malcontent_timer1_extension_agent_request_interface_methods[] =
{
  &org_freedesktop_malcontent_timer1_extension_agent_request_interface__close_method,
  NULL,
};

const GDBusArgInfo org_freedesktop_malcontent_timer1_extension_agent_request_interface__response_signal_args__granted_arg =
{
  -1,  /* ref count */
  (gchar *) "granted",
  (gchar *) "b",
  NULL,  /* no annotations */
};

const GDBusArgInfo org_freedesktop_malcontent_timer1_extension_agent_request_interface__response_signal_args__extra_data_arg =
{
  -1,  /* ref count */
  (gchar *) "extra_data",
  (gchar *) "a{sv}",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const org_freedesktop_malcontent_timer1_extension_agent_request_interface__response_signal_args[] =
{
  &org_freedesktop_malcontent_timer1_extension_agent_request_interface__response_signal_args__granted_arg,
  &org_freedesktop_malcontent_timer1_extension_agent_request_interface__response_signal_args__extra_data_arg,
  NULL,
};

const GDBusSignalInfo org_freedesktop_malcontent_timer1_extension_agent_request_interface__response_signal =
{
  -1,  /* ref count */
  (gchar *) "Response",
  (GDBusArgInfo **) org_freedesktop_malcontent_timer1_extension_agent_request_interface__response_signal_args,
  NULL,  /* no annotations */
};

const GDBusSignalInfo * const org_freedesktop_malcontent_timer1_extension_agent_request_interface_signals[] =
{
  &org_freedesktop_malcontent_timer1_extension_agent_request_interface__response_signal,
  NULL,
};

const GDBusInterfaceInfo org_freedesktop_malcontent_timer1_extension_agent_request_interface =
{
  -1,  /* ref count */
  (gchar *) "org.freedesktop.MalcontentTimer1.ExtensionAgent.Request",
  (GDBusMethodInfo **) org_freedesktop_malcontent_timer1_extension_agent_request_interface_methods,
  (GDBusSignalInfo **) org_freedesktop_malcontent_timer1_extension_agent_request_interface_signals,
NULL,  /* no properties */
  NULL,  /* no annotations */
};

