/*
 * This file is generated by gdbus-codegen, do not modify it.
 *
 * The license of this code is the same as for the D-Bus interface description
 * it was derived from. Note that it links to GLib, so must comply with the
 * LGPL linking clauses.
 */

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include "filtering-iface.h"

#include <string.h>

/* ------------------------------------------------------------------------ */
/* Definitions for org.freedesktop.MalcontentWeb1.Filtering */

const GDBusArgInfo org_freedesktop_malcontent_web1_filtering_interface__update_filters_method_in_args__uid_arg =
{
  -1,  /* ref count */
  (gchar *) "uid",
  (gchar *) "u",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const org_freedesktop_malcontent_web1_filtering_interface__update_filters_method_in_args[] =
{
  &org_freedesktop_malcontent_web1_filtering_interface__update_filters_method_in_args__uid_arg,
  NULL,
};

const GDBusMethodInfo org_freedesktop_malcontent_web1_filtering_interface__update_filters_method =
{
  -1,  /* ref count */
  (gchar *) "UpdateFilters",
  (GDBusArgInfo **) org_freedesktop_malcontent_web1_filtering_interface__update_filters_method_in_args,
  NULL,  /* no out args */
  NULL,  /* no annotations */
};

const GDBusMethodInfo * const org_freedesktop_malcontent_web1_filtering_interface_methods[] =
{
  &org_freedesktop_malcontent_web1_filtering_interface__update_filters_method,
  NULL,
};

const GDBusInterfaceInfo org_freedesktop_malcontent_web1_filtering_interface =
{
  -1,  /* ref count */
  (gchar *) "org.freedesktop.MalcontentWeb1.Filtering",
  (GDBusMethodInfo **) org_freedesktop_malcontent_web1_filtering_interface_methods,
  NULL,  /* no signals */
NULL,  /* no properties */
  NULL,  /* no annotations */
};

