/*
 * This file is generated by gdbus-codegen, do not modify it.
 *
 * The license of this code is the same as for the D-Bus interface description
 * it was derived from. Note that it links to GLib, so must comply with the
 * LGPL linking clauses.
 */

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include "properties-iface.h"

#include <string.h>

/* ------------------------------------------------------------------------ */
/* Definitions for org.freedesktop.DBus.Properties */

const GDBusArgInfo org_freedesktop_dbus_properties_interface__get_method_in_args__interface_name_arg =
{
  -1,  /* ref count */
  (gchar *) "interface_name",
  (gchar *) "s",
  NULL,  /* no annotations */
};

const GDBusArgInfo org_freedesktop_dbus_properties_interface__get_method_in_args__property_name_arg =
{
  -1,  /* ref count */
  (gchar *) "property_name",
  (gchar *) "s",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const org_freedesktop_dbus_properties_interface__get_method_in_args[] =
{
  &org_freedesktop_dbus_properties_interface__get_method_in_args__interface_name_arg,
  &org_freedesktop_dbus_properties_interface__get_method_in_args__property_name_arg,
  NULL,
};

const GDBusArgInfo org_freedesktop_dbus_properties_interface__get_method_out_args__value_arg =
{
  -1,  /* ref count */
  (gchar *) "value",
  (gchar *) "v",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const org_freedesktop_dbus_properties_interface__get_method_out_args[] =
{
  &org_freedesktop_dbus_properties_interface__get_method_out_args__value_arg,
  NULL,
};

const GDBusMethodInfo org_freedesktop_dbus_properties_interface__get_method =
{
  -1,  /* ref count */
  (gchar *) "Get",
  (GDBusArgInfo **) org_freedesktop_dbus_properties_interface__get_method_in_args,
  (GDBusArgInfo **) org_freedesktop_dbus_properties_interface__get_method_out_args,
  NULL,  /* no annotations */
};

const GDBusArgInfo org_freedesktop_dbus_properties_interface__get_all_method_in_args__interface_name_arg =
{
  -1,  /* ref count */
  (gchar *) "interface_name",
  (gchar *) "s",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const org_freedesktop_dbus_properties_interface__get_all_method_in_args[] =
{
  &org_freedesktop_dbus_properties_interface__get_all_method_in_args__interface_name_arg,
  NULL,
};

const GDBusArgInfo org_freedesktop_dbus_properties_interface__get_all_method_out_args__properties_arg =
{
  -1,  /* ref count */
  (gchar *) "properties",
  (gchar *) "a{sv}",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const org_freedesktop_dbus_properties_interface__get_all_method_out_args[] =
{
  &org_freedesktop_dbus_properties_interface__get_all_method_out_args__properties_arg,
  NULL,
};

const GDBusMethodInfo org_freedesktop_dbus_properties_interface__get_all_method =
{
  -1,  /* ref count */
  (gchar *) "GetAll",
  (GDBusArgInfo **) org_freedesktop_dbus_properties_interface__get_all_method_in_args,
  (GDBusArgInfo **) org_freedesktop_dbus_properties_interface__get_all_method_out_args,
  NULL,  /* no annotations */
};

const GDBusArgInfo org_freedesktop_dbus_properties_interface__set_method_in_args__interface_name_arg =
{
  -1,  /* ref count */
  (gchar *) "interface_name",
  (gchar *) "s",
  NULL,  /* no annotations */
};

const GDBusArgInfo org_freedesktop_dbus_properties_interface__set_method_in_args__property_name_arg =
{
  -1,  /* ref count */
  (gchar *) "property_name",
  (gchar *) "s",
  NULL,  /* no annotations */
};

const GDBusArgInfo org_freedesktop_dbus_properties_interface__set_method_in_args__value_arg =
{
  -1,  /* ref count */
  (gchar *) "value",
  (gchar *) "v",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const org_freedesktop_dbus_properties_interface__set_method_in_args[] =
{
  &org_freedesktop_dbus_properties_interface__set_method_in_args__interface_name_arg,
  &org_freedesktop_dbus_properties_interface__set_method_in_args__property_name_arg,
  &org_freedesktop_dbus_properties_interface__set_method_in_args__value_arg,
  NULL,
};

const GDBusMethodInfo org_freedesktop_dbus_properties_interface__set_method =
{
  -1,  /* ref count */
  (gchar *) "Set",
  (GDBusArgInfo **) org_freedesktop_dbus_properties_interface__set_method_in_args,
  NULL,  /* no out args */
  NULL,  /* no annotations */
};

const GDBusMethodInfo * const org_freedesktop_dbus_properties_interface_methods[] =
{
  &org_freedesktop_dbus_properties_interface__get_method,
  &org_freedesktop_dbus_properties_interface__get_all_method,
  &org_freedesktop_dbus_properties_interface__set_method,
  NULL,
};

const GDBusArgInfo org_freedesktop_dbus_properties_interface__properties_changed_signal_args__interface_name_arg =
{
  -1,  /* ref count */
  (gchar *) "interface_name",
  (gchar *) "s",
  NULL,  /* no annotations */
};

const GDBusArgInfo org_freedesktop_dbus_properties_interface__properties_changed_signal_args__changed_properties_arg =
{
  -1,  /* ref count */
  (gchar *) "changed_properties",
  (gchar *) "a{sv}",
  NULL,  /* no annotations */
};

const GDBusArgInfo org_freedesktop_dbus_properties_interface__properties_changed_signal_args__invalidated_properties_arg =
{
  -1,  /* ref count */
  (gchar *) "invalidated_properties",
  (gchar *) "as",
  NULL,  /* no annotations */
};

const GDBusArgInfo * const org_freedesktop_dbus_properties_interface__properties_changed_signal_args[] =
{
  &org_freedesktop_dbus_properties_interface__properties_changed_signal_args__interface_name_arg,
  &org_freedesktop_dbus_properties_interface__properties_changed_signal_args__changed_properties_arg,
  &org_freedesktop_dbus_properties_interface__properties_changed_signal_args__invalidated_properties_arg,
  NULL,
};

const GDBusSignalInfo org_freedesktop_dbus_properties_interface__properties_changed_signal =
{
  -1,  /* ref count */
  (gchar *) "PropertiesChanged",
  (GDBusArgInfo **) org_freedesktop_dbus_properties_interface__properties_changed_signal_args,
  NULL,  /* no annotations */
};

const GDBusSignalInfo * const org_freedesktop_dbus_properties_interface_signals[] =
{
  &org_freedesktop_dbus_properties_interface__properties_changed_signal,
  NULL,
};

const GDBusInterfaceInfo org_freedesktop_dbus_properties_interface =
{
  -1,  /* ref count */
  (gchar *) "org.freedesktop.DBus.Properties",
  (GDBusMethodInfo **) org_freedesktop_dbus_properties_interface_methods,
  (GDBusSignalInfo **) org_freedesktop_dbus_properties_interface_signals,
NULL,  /* no properties */
  NULL,  /* no annotations */
};

