/* -*- mode: C; c-file-style: "gnu"; indent-tabs-mode: nil; -*-
 *
 * Copyright 2025 GNOME Foundation, Inc.
 *
 * SPDX-License-Identifier: GPL-2.0-or-later
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 *
 * Authors:
 *  - Ignacy Kuchciński <ignacykuchcinski@gnome.org>
 */

#pragma once

#include <adwaita.h>
#include "cc-screen-time-statistics-row.h"

G_BEGIN_DECLS

#define MCT_TYPE_SCREEN_TIME_STATISTICS_ROW (mct_screen_time_statistics_row_get_type ())
G_DECLARE_FINAL_TYPE (MctScreenTimeStatisticsRow, mct_screen_time_statistics_row, MCT, SCREEN_TIME_STATISTICS_ROW, CcScreenTimeStatisticsRow)

MctScreenTimeStatisticsRow *mct_screen_time_statistics_row_new (GDBusConnection *connection,
                                                                uid_t            uid);

G_END_DECLS
